@@ -2,6 +2,8 @@ from functools import reduce
 import operator
 import os
 import sys
+import logging
+
 import numpy as np
 
 import dask.base
@@ -17,10 +19,12 @@ from .docstrings import docsubst
 import vaex.utils
 
 
+logger = logging.getLogger("vaex.agg")
+
 if vaex.utils.has_c_extension:
     import vaex.superagg
 
-
+_min = min  # we're gonna overwrite builtin min
 aggregates = {}
 
 
@@ -270,22 +274,44 @@ class AggregatorDescriptorBasic(AggregatorDescriptor):
             return self.finish(value)
         return [task], finish(task)
 
-    def _create_operation(self, grid):
+    def _create_operation(self, grid, nthreads):
         agg_op_type = vaex.utils.find_type_from_dtype(vaex.superagg, self.name + "_", self.dtype_in)
         bytes_per_cell = self.dtype_out.numpy.itemsize
         cells = reduce(operator.mul, [len(binner) for binner in grid.binners], 1)
-        predicted_memory_usage = bytes_per_cell * cells
+        grids = nthreads
+        ncells = len(grid)
+        # the more grid cells we have, the less more work have to do to
+        # merge/reduce the grids
+        if ncells >= 1e4:
+            grids = _min(32, nthreads)
+        if ncells >= 1e5:
+            grids = _min(16, nthreads)
+        if ncells >= 1e6:
+            grids = _min(8, nthreads)
+        if grids < 1:
+            grids = 1
+        if logger.isEnabledFor(logging.INFO):
+            logger.info("Using %r grids for %r thread for aggerator %r (total grid cells %s)", grids, nthreads, self, f"{ncells:,}")
+        predicted_memory_usage = bytes_per_cell * cells * grids
         vaex.memory.local.agg.pre_alloc(predicted_memory_usage, f"aggregator data for {agg_op_type}")
-        agg_op = agg_op_type(grid, *self.agg_args)
+        agg_op = agg_op_type(grid, grids, nthreads, *self.agg_args)
         used_memory = sys.getsizeof(agg_op)
         if predicted_memory_usage != used_memory:
             raise RuntimeError(f'Wrong prediction for {agg_op_type}, expected to take {predicted_memory_usage} bytes but actually used {used_memory}')
         return agg_op
 
     def get_result(self, agg_operation):
-        grid = np.asarray(agg_operation)
+        grid = agg_operation.get_result()
         if not self.edges:
-            grid = vaex.utils.extract_central_part(grid)
+            def binner2slice(binner):
+                if 'BinnerScalar_' in str(binner):
+                    return slice(2, -1)
+                elif 'BinnerOrdinal_' in str(binner):
+                    return slice(0, -2)
+                else:
+                    raise TypeError(f'Binner not supported with edges=False {binner}')
+            slices = [binner2slice(binner) for binner in agg_operation.grid.binners]
+            grid = grid[tuple(slices)]
         return grid
 
 
@@ -307,12 +333,15 @@ class AggregatorDescriptorNUnique(AggregatorDescriptorBasic):
         super()._prepare_types(df)
         self.dtype_out = DataType(np.dtype('int64'))
 
-    def _create_operation(self, grid):
+    def _create_operation(self, grid, nthreads):
+        grids = 1  # this is using a shared hashmap, which is thread safe
         agg_op_type = vaex.utils.find_type_from_dtype(vaex.superagg, self.name + "_", self.dtype_in)
         cells = reduce(operator.mul, [len(binner) for binner in grid.binners], 1)
-        predicted_memory_usage = self.dtype_out.numpy.itemsize * cells
+        grid0 = vaex.superagg.Grid([])
+        agg_op_test = agg_op_type(grid0, grids, nthreads, self.dropmissing, self.dropnan)
+        predicted_memory_usage = sys.getsizeof(agg_op_test) * cells
         vaex.memory.local.agg.pre_alloc(predicted_memory_usage, f"aggregator data for {agg_op_type}")
-        agg_op = agg_op_type(grid, self.dropmissing, self.dropnan)
+        agg_op = agg_op_type(grid, grids, nthreads, self.dropmissing, self.dropnan)
         used_memory = sys.getsizeof(agg_op)
         if predicted_memory_usage != used_memory:
             raise RuntimeError(f'Wrong prediction for {agg_op_type}, expected to take {predicted_memory_usage} bytes but actually used {used_memory}')
